// src/backend/text_simulation.hh - Copyright 2005, 2006, University
//               of Padova, dept. of Pure and Applied
//               Mathematics
//
// This file is part of SGPEMv2.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SGPEMv2 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SGPEMv2; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#ifndef TEXT_SIMULATION_HH
#define TEXT_SIMULATION_HH 1

#include "config.h"
#include "gettext.h"

#include <sgpemv2/simulation.hh>
#include <sgpemv2/simulation_observer.hh>
#include <sgpemv2/string_utils.hh>

#include <glibmm/thread.h>
#include <glibmm/ustring.h>
#include <utility>
#include <map>

namespace sgpem
{
  template <typename T>
  class CommandParameter;
}

namespace sgpem
{
  class TextSimulation;

  /**
    \brief Textual interface for the simulation controller.

    A command-based interface is used, so methods of the base class can be called with
    a proper command string obtained from the input device(s).
   */
  class SG_DLLEXPORT TextSimulation : public SimulationObserver
  {
  public:
    TextSimulation();
    ~TextSimulation();

    /**
      \brief Executes a command read from an input device.

      A list of supported commands can be obtained with the \b help command.
     */
    static void parse_command(TextSimulation &sim, const Glib::ustring& str);

  private:
    /**
      Prints the actual state of the simulation, with emphasis on the current
      status of the scheduling process (ready queue and running process).
     */
    virtual void update(const Simulation& changed_simulation);

    bool check_arguments_num(const Tokens& arguments, unsigned int num);
    bool unsaved_ask_confirm() const;
    template <typename Container>
    void show(const Container& entities);
    template <typename T>
    void get_parameter(CommandParameter<T>& parameter);
    template <typename PolicyType>
    void configure_policy(PolicyType& policy);
    template <typename GatekeeperType>
    void show_policies(const Tokens& arguments);
    
    void on_run(const Tokens& arguments);
    void on_pause(const Tokens& arguments);
    void on_jumpto(const Tokens& arguments);
    void on_stop(const Tokens& arguments);
    void on_configure(const Tokens& arguments);
    void on_help(const Tokens& arguments);
    void on_quit(const Tokens& arguments);
    void on_get(const Tokens& arguments);
    void on_set(const Tokens& arguments);
    void on_show(const Tokens& arguments);
    void on_show_processes(const Tokens& arguments);
    void on_show_resources(const Tokens& arguments);
    void on_show_threads(const Tokens& arguments);
    void on_show_requests(const Tokens& arguments);
    void on_show_subrequests(const Tokens& arguments);
    void on_show_cpu_policies(const Tokens& arguments);
    void on_show_resource_policies(const Tokens& arguments);
    void on_add(const Tokens& arguments);
    void on_add_process(const Tokens& arguments);
    void on_add_resource(const Tokens& arguments);
    void on_add_thread(const Tokens& arguments);
    void on_add_request(const Tokens& arguments);
    void on_add_subrequest(const Tokens& arguments);
    void on_remove(const Tokens& arguments);
    void on_remove_process(const Tokens& arguments);
    void on_remove_resource(const Tokens& arguments);
    void on_remove_thread(const Tokens& arguments);
    void on_remove_request(const Tokens& arguments);
    void on_remove_subrequest(const Tokens& arguments);
    void on_save(const Tokens& arguments);
    void on_load(const Tokens& arguments);

    // FIXME This is a temporary replacement for the
    // to-be written I/O layer
    static void p_stdout(const Glib::ustring& str);
    static void p_stderr(const Glib::ustring& str);
    static Glib::ustring readline();

    bool _saved;
  };

}


#endif
